package cs451.net.handler;

import cs451.net.event.*;
import cs451.net.event.message.NetMessageAbstract;
import cs451.parser.Host;
import cs451.parser.Parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * NetEventHandler for First In First Out
 *
 * DeliverIf: packet from queue satisfies FIFO ordering
 *
 * Send: sends next packet of current host
 * Deliver: Add packet to received packet queue for DeliverIf
 *
 * @author C. Hölzl
 */
public class NetHandlerFIFO extends NetEventHandlerAbstract {

    private static final ConcurrentHashMap<Integer,AtomicInteger> rsn = new ConcurrentHashMap<>();
    private static final ConcurrentHashMap<Integer,PriorityBlockingQueue<NetMessageAbstract>> pending = new ConcurrentHashMap<>();

    public NetHandlerFIFO(Class<? extends NetEventHandlerAbstract> deliverLayer, Class<? extends NetEventHandlerAbstract> broadcastLayer) {
        super(deliverLayer,broadcastLayer);
    }

    @Override
    public synchronized void deliverIf(){
        pending.forEach((hid, mq) -> {
            AtomicInteger crsn = rsn.get(hid);
            if(crsn == null) return;
            Collection<NetMessageAbstract> msgc = new ArrayList<>();
            mq.drainTo(msgc);
            msgc.forEach(msg ->
                crsn.getAndUpdate(v -> {
                    if (msg.id.equals(v)){
                        deliverNextSync(NetEvent.Message(msg));
                        return v+1;
                    }else{
                        mq.add(msg);
                        return v;
                    }
                })
            );
        });
    }

    @Override
    public void deliver(NetEvent ne) {
        pending.computeIfPresent(ne.message.src, (h, mq) -> {
            mq.add(ne.message);
            requestCondBD(true);
            return mq;
        });
    }

    @Override
    public boolean isDone() {
        return pending.values().stream().allMatch(hmp -> hmp.stream().noneMatch(msg-> msg.id.equals(rsn.getOrDefault(msg.src, new AtomicInteger(1)).get())));
    }

    @Override
    public void start(Host h, Parser p) {
        for (Host host : p.hosts()) {
            pending.put(host.getId(),new PriorityBlockingQueue<>());
            rsn.put(host.getId(), new AtomicInteger(1));
        }
    }
}
