package cs451.net.handler;

import cs451.net.NetManager;
import cs451.net.event.*;
import cs451.net.event.message.NetMessageAbstract;
import cs451.tools.Pair;

import java.util.*;
import java.util.concurrent.*;

/**
 * NetEventHandler for Stuborn Link
 *
 * SendIf: Sends a message again if not Acknowledged
 *
 * Send: Add Message to "Awaiting Acknowledge" queue
 * Deliver: delivers to next layer (Drops ACK messages and removes peer from "Awaiting Acknowledge"
 * Crash: Mark peer as timed out to stop further sends/acknowledgements
 *
 * @author C. Hölzl
 */
public class NetHandlerSL extends NetEventHandlerAbstract {

    private static final Set<Integer> hasTimeout = ConcurrentHashMap.newKeySet();

    private static final PriorityBlockingQueue<Pair<Integer, NetMessageAbstract>> sending = new PriorityBlockingQueue<>();

    public NetHandlerSL(Class<? extends NetEventHandlerAbstract> deliverLayer, Class<? extends NetEventHandlerAbstract> broadcastLayer) {
        super(deliverLayer,broadcastLayer);
    }

    @Override
    public synchronized void beat(){
        sending.removeIf(ppm -> hasTimeout.contains(ppm.first()));
        sending.stream().limit(NetManager.WINDOW_WIDTH).forEach(ppm -> NetManager.hosts.stream().filter(hl->hl.getId()==(ppm.first())).findFirst()
                .ifPresent(host -> broadcastNextSync(NetEvent.Message(host, ppm.second()))));
    }

    @Override
    public void broadcast(NetEvent ne) {
        if (!hasTimeout.contains(ne.peer.getId()))
            sending.add(new Pair<>(ne.peer.getId(), ne.message));
        broadcastNextSync(ne);
    }

    @Override
    public void deliver(NetEvent ne) {
        switch (ne.message.tpe){
            case ACK:
                sending.removeIf(ppm -> (ne.peer.getId()==ppm.first() &&
                        ne.message.equals(ppm.second())));
                break;
            case DATA:
                broadcastNextSync(NetEvent.MessageACK(ne.peer,ne.message));
                deliverNextSync(ne);
                break;
            default:
                break;
        }
    }

    @Override
    public void crash(NetEvent ne) {
        hasTimeout.add(ne.peer.getId());
        crashNextSync(ne);
    }

    @Override
    public boolean isDone() {
        return sending.isEmpty();
    }
}
