package cs451.net.handler;

import cs451.net.event.NetEvent;
import cs451.net.event.message.NetMessageAbstract;
import cs451.parser.Host;
import cs451.parser.Parser;
import cs451.tools.Pair;

import java.util.Map;
import java.util.Set;
import java.util.concurrent.*;

/**
 * NetEventHandler for Uniform Reliable Broadcast
 *
 * DeliverIf: pending has been acked by all correct processes (and has not been delivered)
 *
 * Send: Add to pending and send
 * Deliver: Ack and Add to pending
 * Crash: Remove host from correct processes
 *
 * @author C. Hölzl
 */
public class NetHandlerURB extends NetEventHandlerAbstract {

    private static final Set<Integer> correct = ConcurrentHashMap.newKeySet();
    private static final Map<Pair<Integer,Integer>, Set<Integer>> ack = new ConcurrentHashMap<>();
    private static final Set<Pair<Integer, Integer>> delivered = ConcurrentHashMap.newKeySet();
    private static final Map<Pair<Integer, Integer>, NetMessageAbstract> pending = new ConcurrentHashMap<>();

    public NetHandlerURB(Class<? extends NetEventHandlerAbstract> deliverLayer, Class<? extends NetEventHandlerAbstract> broadcastLayer) {
        super(deliverLayer,broadcastLayer);
    }

    @Override
    public synchronized void deliverIf(){
        pending.keySet().stream().filter(smp ->
             ack.getOrDefault(smp,ConcurrentHashMap.newKeySet()).containsAll(correct)).
                filter(delivered::add).
                forEach(smp -> deliverNextAsync(NetEvent.Message(pending.get(smp))));
    }

    @Override
    public void broadcast(NetEvent ne) {
        pending.put(new Pair<>(ne.message.src, ne.message.id),NetMessageAbstract.MSG(ne.message));
        requestCondBD(true);
        broadcastNextSync(ne);
    }

    @Override
    public void deliver(NetEvent ne) {
        Pair<Integer,Integer> smp = new Pair<>(ne.message.src,ne.message.id);
        ack.compute(smp,(k, v) -> {
            if(v == null){
                v = ConcurrentHashMap.newKeySet();
            }
            v.add(ne.peer.getId());
            return v;
        });
        if(pending.putIfAbsent(smp, NetMessageAbstract.MSG(ne.message))==null){
            broadcastNextAsync(ne);
        }
        requestCondBD(true);
    }

    @Override
    public void crash(NetEvent ne) {
        correct.remove(ne.peer.getId());
        requestCondBD(true);
    }

    @Override
    public boolean isDone() {
        return delivered.containsAll(pending.keySet());
    }

    @Override
    public void start(Host h, Parser p) {
        p.hosts().forEach(ch-> correct.add(ch.getId()));
    }
}